using System;
using System.Collections.Generic;
using System.Text;

namespace HeroSquareRoots
{
    public class CalcProgressArgs : EventArgs
    {
        private readonly HeroStatus currentStatus;
        private bool cancel;
        private readonly bool done;

        public bool Done
        {
            get { return done; }
        }

        public HeroStatus CurrentStatus
        {
            get { return currentStatus; }
        }

        public bool Cancel
        {
            get { return cancel; }
            set { cancel = value; }
        }

        public CalcProgressArgs(HeroStatus status, bool finished)
        {
            currentStatus = status;
            done = finished;
        }
    }

    public class HeroCalc
    {
        public event EventHandler<CalcProgressArgs> OnUpdateProgress;
        private HeroStatus currentStatus;

        public HeroStatus CurrentStatus
        {
            get { return currentStatus; }
        }

        private List<double> epsilonHistory = new List<double>();

        private object synchHandle;

        private object GetSynchHandle()
        {
            System.Threading.Interlocked.CompareExchange(ref synchHandle, new object(), null);
            return synchHandle;
        }

        public double[] GetEpsilon(int numValues)
        {
            lock (GetSynchHandle())
            {
                // return the last N items in the epsilon array.
                int returnedItems = Math.Min(numValues, epsilonHistory.Count);
                double[] rVals = new double[returnedItems];
                int index = epsilonHistory.Count - returnedItems;
                epsilonHistory.CopyTo(index, rVals, 0, returnedItems);
            }
            return rVals;
        }

        public HeroCalc()
        {
        }

        public void DoCalc(double square, double epsilon)
        {
            if (epsilon < 0)
                throw new ArgumentException("tolerance must be a postive number");
            currentStatus = new HeroStatus(square, 1, 0);
            lock (GetSynchHandle())
            {
                epsilonHistory.Clear();
                epsilonHistory.Add(currentStatus.Epsilon);
            }

            if (raiseProgress(false))
                return;
            while (currentStatus.Epsilon > epsilon)
            {
                nextIteration();
                lock (GetSynchHandle())
                {
                    epsilonHistory.Add(currentStatus.Epsilon);
                }
                    if (raiseProgress(false))
                        return;
            }
            raiseProgress(true);
        }

        private void nextIteration()
        {
            double nextGuess =
                (currentStatus.Operand / currentStatus.Guess + currentStatus.Guess)
                 / 2.0;
            HeroStatus updatedStatus = new HeroStatus(currentStatus.Operand, nextGuess,
                currentStatus.NumIterations + 1);
            System.Threading.Interlocked.Exchange(ref currentStatus, updatedStatus);
        }

        private bool raiseProgress(bool done)
        {
            if (OnUpdateProgress != null)
            {
                CalcProgressArgs args = new CalcProgressArgs(currentStatus, done);
                OnUpdateProgress(this, args);
                if (args.Cancel)
                    return true;
            }
            return false;
        }
    }
}
